/*
 * save the original functions of HistoryMenu
 */
HistorySubmenus2.HM = {
	HM_placeProperty: Object.getOwnPropertyDescriptor(HistoryMenu.prototype, "place"),
	HM_rebuildPopup: HistoryMenu.prototype._rebuildPopup,
	HM_ensureMarkers: HistoryMenu.prototype._ensureMarkers,
	HM_uninit: HistoryMenu.prototype.uninit,

	get placeProperty()
		this.HM_placeProperty ||
				Object.getOwnPropertyDescriptor(PlacesMenu.prototype, "place") ||
				Object.getOwnPropertyDescriptor(PlacesViewBase.prototype, "place"),
	get rebuildPopup()
		this.HM_rebuildPopup || PlacesMenu.prototype._rebuildPopup,
	get ensureMarkers()
		this.HM_ensureMarkers || PlacesMenu.prototype._ensureMarkers,
	get uninit()
		this.HM_uninit || PlacesMenu.prototype.uninit
};

/*
 * save the original functions of BookmarksEventHandler
 */
HistorySubmenus2.BEH = {
	fillInBHTooltip: BookmarksEventHandler.fillInBHTooltip
};

/*
 * wrap the fillInBHTooltip to fill the browse time of history item
 */
BookmarksEventHandler.fillInBHTooltip = function BEH_fillInBHTooltip(aDocument, aEvent) {
	let tooltipTime = aDocument.getElementById("bhtTimeText");

	if (tooltipTime) {
		let tooltipNode = aDocument.tooltipNode;
		let node = tooltipNode && tooltipNode._placesNode;
		if (node && node.bookmarkIndex == -1) {
			const dateService = HistorySubmenus2.dateService;

			let date = new Date(node.time / 1000);
			tooltipTime.value = dateService.FormatDateTime(
					HistorySubmenus2.locale, dateService.dateFormatShort, dateService.timeFormatSeconds,
					date.getFullYear(), date.getMonth() + 1, date.getDate(),
					date.getHours(), date.getMinutes(), date.getSeconds());
			tooltipTime.hidden = false;
		} else
			tooltipTime.hidden = true;
	}

	return HistorySubmenus2.BEH.fillInBHTooltip.apply(this, arguments);
};

/*
 * wrap the setter of property "place"
 * since the places uri is hard coded in constructor,
 * so we intercept the places uri here
 */
HistoryMenu.prototype.__defineSetter__("place", function(val) {
	let maxResults = HistorySubmenus2.prefService.getIntPref("historyCount");

	/*
	 * if maxResults = 0, the query will return all results but not none
	 * so we need to make a little trick for showing zero result
	 */
	this._showZeroResult = !maxResults;
	if (!maxResults)
		maxResults = 1;

	/*
	 * update the max results property
	 */
	let currentMaxResults = /\d+/.exec(/maxResults=\d+/.exec(val));
	if (currentMaxResults != maxResults)
		val = val.replace("maxResults=" + currentMaxResults, "maxResults=" + maxResults);

	HistorySubmenus2.HM.placeProperty.set.call(this, val);
});

HistoryMenu.prototype.__defineGetter__("place", function() {
	return HistorySubmenus2.HM.placeProperty.get.call(this);
});

/*
 * properties
 */
HistoryMenu.prototype._subBuiltDate = null;
HistoryMenu.prototype._subResult = null;
HistoryMenu.prototype._subMenus = [];
HistoryMenu.prototype._showZeroResult = false;

/*
 * wrap the _rebuildPopup
 * we do some preparation for sub-menus after building top-menu
 */
HistoryMenu.prototype._rebuildPopup = function HM_rebuildPopup(popup) {
	const todayRef = Components.interfaces.nsINavHistoryQuery.TIME_RELATIVE_TODAY;
	const maxCount = HistorySubmenus2.prefService.getIntPref("historyCount");
	const submenuCount = HistorySubmenus2.prefService.getIntPref("submenuCount");
	const showScrollbar = HistorySubmenus2.prefService.getBoolPref("showScrollbar");

	/*
	 * call the original _rebuildPopup or the default one
	 */
	let result = HistorySubmenus2.HM.rebuildPopup.apply(this, arguments);

	/*
	 * build the sub-menus when history menu is first open or when the day has passed
	 * well, here we only check the date and ignore the month and year
	 */
	 if (!this._subBuiltDate || this._subBuiltDate.getDate() != new Date().getDate()) {
	 	this._rebuildSubMenus(popup);
 	 	this._subBuiltDate = new Date();
	 }

	/*
	 * prepare for sub-menus
	 */
	let hasSubShow = false;
	if (submenuCount) {
		/*
		 * create the result of days ago
		 */
		if (!this._subResult) {
			let history = PlacesUtils.history;
			let queries = {};
			let options = {};
			history.queryStringToQueries(
					"redirectsMode=2&sort=4" +
							"&endTimeRef=" + todayRef +
							"&endTime=" + (24 * 60 * 60 * 1000 * 1000) +
							"&beginTimeRef=" + todayRef +
							"&beginTime=" + (- submenuCount * 24 * 60 * 60 * 1000 * 1000),
					queries, {}, options);
			this._subResult = history.executeQueries(queries.value, queries.value.length,
					options.value);

			/*
			 * we keep the result open since it takes 0.5s for opening each time
			 */
			this._subResult.root.containerOpen = true;
		}

		/*
		 * prepare the result for each sub-menus and hide the empty ones
		 */
		let resultRoot = this._subResult.root;
		let childCount = resultRoot.childCount;
		let currentChildIndex = maxCount;

		for (let i in this._subMenus) {
			let submenu = this._subMenus[i];
			let subpopup = submenu.lastChild;
			let dateOfSub = HistorySubmenus2.getStartOfDate(
					new Date(new Date() - subpopup._age * 24 * 60 * 60 * 1000)).getTime();
			subpopup._result = [];

			/*
			 * assign the result for each sub-menu
			 */
			while (currentChildIndex < childCount) {
				let child = resultRoot.getChild(currentChildIndex);
				let dateOfChild = HistorySubmenus2.getStartOfDate(new Date(child.time / 1000)).getTime();

				/*
				 * we use the integer time compare here
				 * since the date object compare is incorrect in this case, weird
				 */
				if (dateOfChild == dateOfSub)
					subpopup._result.push(child);
				else if (dateOfChild < dateOfSub)
					break;

				currentChildIndex++;
			}

			if (!subpopup._result.length)
				submenu.hidden = true;
			else {
				submenu.hidden = false;
				hasSubShow = true;

				//update the open sub-menus
				if (submenu.open)
					submenu._placesView._rebuildPopup(subpopup);
			}
		}

	}
	popup._endMarker.hidden = !hasSubShow;

	/*
	 * a little trick, remove the additional item here
	 * please read HistoryMenu.prototype.place
	 */
	if (this._showZeroResult) {
		this._setEmptyPopupStatus(popup, true);
		let lastHistory = popup._startMarker.nextSibling;
		if (lastHistory._placesNode)
			this._removeChild(lastHistory);
	}
	
	/*
	 * update the properties of history menu
	 */
	popup.setAttribute("historypopup", true);
	popup.setAttribute("showscrollbar", showScrollbar);
	
	/*
	 * return the result of original function
	 */
	return result;
};

/*
 * wrap the _ensureMarkers
 * we add one more markers for history menu: sub-menus marker
 */
HistoryMenu.prototype._ensureMarkers = function HM_ensureMarkers(popup) {
	/*
	 * call the original _ensureMarkers or the default one
	 */
	let result = HistorySubmenus2.HM.ensureMarkers.apply(this, arguments);
	
	/*
	 * we do our jobs after the preparing of markers
	 */
	if (!popup._endSubMarker) {
		let els = popup.getElementsByClassName("endHistorySubmenusSeparator");
		for (let i = 0; i < els.length; i++)
			if (els[i].parentNode == popup) {
				popup._endSubMarker = els[i];
				break;
			}
			
		if (!popup._endSubMarker) {
			popup._endSubMarker = popup.insertBefore(document.createElement("menuseparator"),
					popup._endMarker.nextSibling);
			popup._endSubMarker.className = "endHistorySubmenusSeparator";
		}
	}
	
	/*
	 * return the result of original function
	 */
	return result;
};

/*
 * add a new function for building sub-menus
 */
HistoryMenu.prototype._rebuildSubMenus =  function HM_rebuildSubMenus(popup) {
	/*
	 * remove the old ones
	 */
	this._cleanSubMenus(popup);

	/*
	 * create submenus
	 */
	const submenuCount = HistorySubmenus2.prefService.getIntPref("submenuCount");
	const dateFormat = HistorySubmenus2.prefService.getComplexValue("dateFormat",
			Components.interfaces.nsISupportsString).data;

	for (let age = 0; age < submenuCount; age++) {
		let submenu = popup.insertBefore(document.createElement("menu"), popup._endSubMarker);
		submenu.className = "history-submenu";
		submenu.setAttribute("label", HistorySubmenus2.formatDate(
				new Date(new Date() - age * 24 * 60 * 60 * 1000), dateFormat));

		/*
		 * create popup
		 */
		let subpopup = submenu.appendChild(document.createElement("menupopup"));
		subpopup._age = age;
		subpopup.setAttribute("placespopup", true);
		subpopup.setAttribute("onpopupshowing",
				"if (!this.parentNode._placesView) new HistorySubMenu(event);");

		this._subMenus.push(submenu);
	}
};

HistoryMenu.prototype._cleanSubMenus = function HM_cleanSubMenus(popup) {
	if (this._subMenus)
		for (let i in this._subMenus) {
			let menu = this._subMenus[i];
			if (menu._placesView)
				menu._placesView.uninit();
			popup.removeChild(menu);
		}

	this._subMenus = [];
};

/*
 * wrap the uninit
 */
HistoryMenu.prototype.uninit = function HM_uninit() {
	if (this._subResult) {
		this._subResult.root.containerOpen = false;
		this._subResult = null;
	}
	let popup = this._viewElt.lastChild;
	this._cleanSubMenus(popup);
	popup.removeAttribute("historypopup");
	popup.removeAttribute("showscrollbar");
	
	return HistorySubmenus2.HM.uninit.call(this);
};

/*
 * view for sub-menus
 */
function HistorySubMenu(aPopupShowingEvent) {
	/*
	 * the sub-menus extands PlacesMenu for some functions like drag and drop support
	 */
	this.__proto__.__proto__ = PlacesMenu.prototype;

	/*
	 * since the sub-menus doesn't use query result, so we give some junk places string for
	 * construction. but i think there should be a places uri that is better than the default "place:"...
	 */
	PlacesMenu.call(this, aPopupShowingEvent, "");
}

HistorySubMenu.prototype = {
	_onPopupShowing: function HSM_onPopupShowing(aEvent) {
		this._rebuildPopup(aEvent.originalTarget);
	},

	_rebuildPopup: function HSM_rebuildPopup(popup) {
		this._cleanPopup(popup);

		for (let i in popup._result)
			this._insertNewItemToPopup(popup._result[i], popup);

		this._setEmptyPopupStatus(popup, !popup._result.length);
	}
};